# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re
from typing import Optional

from packaging.version import InvalidVersion

from pontos.git import Git, TagSort
from pontos.version.version import Version, parse_version


def strip_version(version: str) -> str:
    """
    Strips a leading 'v' from a version string

    E.g. v1.2.3 will be converted to 1.2.3
    """
    if version and version[0] == "v":
        return version[1:]

    return version


def check_develop(version: str) -> bool:
    """
    Checks if the given Version is a develop version

    Returns True if yes, False if not
    """
    return True if Version(version).dev is not None else False


def is_version_pep440_compliant(version: str) -> bool:
    """
    Checks if the provided version is a PEP 440 compliant version string
    """
    return version == safe_version(version)


def safe_version(version: str) -> str:
    """
    Returns the version as a string in `PEP440`_ compliant
    format.

    .. _PEP440:
       https://www.python.org/dev/peps/pep-0440
    """
    try:
        return str(Version(version))
    except InvalidVersion:
        version = version.replace(" ", ".")
        return re.sub("[^A-Za-z0-9.]+", "-", version)


def versions_equal(new_version: str, old_version: str) -> bool:
    """
    Checks if new_version and old_version are equal
    """
    return safe_version(old_version) == safe_version(new_version)


def get_last_release_version(
    git_tag_prefix: Optional[str] = "",
    *,
    ignore_pre_releases: Optional[bool] = False,
) -> Optional[Version]:
    """Get the last released Version from git.

    Args:
        git_tag_prefix: Git tag prefix to consider
        ignore_pre_release: Ignore pre releases and only consider non pre
            releases. Default is False.

    Returns:
        Last released git-tag as Version if tags were found
        or None
    """

    tag_list = Git().list_tags(sort=TagSort.VERSION)

    while tag_list:
        last_release_version = tag_list[-1]
        last_release_version = last_release_version.strip(git_tag_prefix)

        version = parse_version(last_release_version)
        if not version.is_prerelease or not ignore_pre_releases:
            return version

        tag_list = tag_list[:-1]

    return None



# Copyright (C) 2023 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Tuple, Type

from pontos.version.errors import ProjectError

from .cmake import CMakeVersionCommand
from .go import GoVersionCommand
from .javascript import JavaScriptVersionCommand
from .python import PythonVersionCommand
from .version import VersionCommand

_COMMANDS: Tuple[Type[VersionCommand]] = (
    CMakeVersionCommand,
    GoVersionCommand,
    JavaScriptVersionCommand,
    PythonVersionCommand,
)


def gather_project() -> VersionCommand:
    """
    Get the fitting VersionCommand for a project in the current working
    directory

    Raises:
        ProjectError if no fitting VersionCommand could be found
    """
    for cmd in _COMMANDS:
        command = cmd()
        if command.project_found():
            return command

    raise ProjectError("No project settings file found")



# Copyright (C) 2023 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse


def initialize_default_parser() -> argparse.ArgumentParser:
    """
    Returns a default argument parser containing:
    - verify
    - show
    - update
    """
    parser = argparse.ArgumentParser(
        description="Version handling utilities.",
        prog="version",
    )
    parser.add_argument(
        "--quiet", help="don't print messages", action="store_true"
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        help="additional help",
        dest="command",
    )

    verify_parser = subparsers.add_parser("verify")
    verify_parser.add_argument("version", help="version string to compare")
    subparsers.add_parser("show")

    update_parser = subparsers.add_parser("update")
    update_parser.add_argument("version", help="version string to use")
    update_parser.add_argument(
        "--force",
        help="don't check if version is already set",
        action="store_true",
    )
    update_parser.add_argument(
        "--develop",
        help="indicates if it is a develop version",
        action="store_true",
    )
    return parser



# Copyright (C) 2023 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from pontos.testing import temp_directory, temp_python_module
from pontos.version.cmake import CMakeVersionCommand
from pontos.version.commands import gather_project
from pontos.version.errors import ProjectError
from pontos.version.go import GoVersionCommand
from pontos.version.javascript import JavaScriptVersionCommand
from pontos.version.python import PythonVersionCommand


class GatherProjectTestCase(unittest.TestCase):
    def test_no_project_found(self):
        with temp_directory(change_into=True), self.assertRaisesRegex(
            ProjectError, "No project settings file found"
        ):
            gather_project()

    def test_python_project(self):
        content = "__version__ = '1.2.3'"
        with temp_python_module(
            content, name="foo", change_into=True
        ) as temp_module:
            temp_dir = temp_module.parent / "pyproject.toml"
            temp_dir.write_text(
                '[tool.poetry]\nversion = "1.2.3"\n'
                '[tool.pontos.version]\nversion-module-file = "foo.py"',
                encoding="utf8",
            )

            self.assertIsInstance(gather_project(), PythonVersionCommand)

    def test_go_project(self):
        with temp_directory(change_into=True) as temp_dir:
            project_file = temp_dir / "go.mod"
            project_file.touch()
            version_file = temp_dir / "version.go"
            version_file.write_text('var version = "1.2.3"')

            self.assertIsInstance(gather_project(), GoVersionCommand)

    def test_javascript_project(self):
        with temp_directory(change_into=True) as temp_dir:
            version_file = temp_dir / "package.json"
            version_file.write_text(
                '{"name": "foo", "version": "1.2.3"}', encoding="utf8"
            )

            self.assertIsInstance(gather_project(), JavaScriptVersionCommand)

    def test_cmake_project_version(self):
        with temp_directory(change_into=True) as temp_dir:
            version_file = temp_dir / "CMakeLists.txt"
            version_file.write_text("project(VERSION 1.2.3)", encoding="utf8")

            self.assertIsInstance(gather_project(), CMakeVersionCommand)



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Literal, Type, Union

from packaging.version import InvalidVersion
from packaging.version import Version as PackagingVersion

from .errors import VersionError


class Version(PackagingVersion):
    """
    A class handling version information
    """


def parse_version(version: str) -> Version:
    try:
        return Version(version)
    except InvalidVersion as e:
        raise VersionError(e) from None


@dataclass
class VersionUpdate:
    previous: Version
    new: Version
    changed_files: list[Path] = field(default_factory=list)


class VersionCalculator:
    def next_patch_version(self, current_version: Version) -> Version:
        """
        Get the next patch version from a valid version

        Examples:
            "1.2.3" will return "1.2.4"
            "1.2.3.dev1" will return "1.2.3"

        Raises:
            VersionError if version is invalid.
        """

        if current_version.is_prerelease:
            next_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro}"
            )
        else:
            next_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro + 1}"
            )

        return next_version

    def next_calendar_version(self, current_version: Version) -> Version:
        """
        Find the correct next calendar version by checking latest version and
        the today's date

        Raises:
            VersionError if version is invalid.
        """

        today = datetime.today()
        current_year_short = today.year % 100

        if current_version.major < current_year_short or (
            current_version.major == current_year_short
            and current_version.minor < today.month
        ):
            return parse_version(f"{current_year_short}.{today.month}.0")

        if (
            current_version.major == today.year % 100
            and current_version.minor == today.month
        ):
            if current_version.dev is None:
                release_version = parse_version(
                    f"{current_year_short}.{today.month}."
                    f"{current_version.micro + 1}"
                )
            else:
                release_version = parse_version(
                    f"{current_year_short}.{today.month}."
                    f"{current_version.micro}"
                )
            return release_version
        else:
            raise VersionError(
                f"'{current_version}' is higher than "
                f"'{current_year_short}.{today.month}'."
            )

    def next_minor_version(self, current_version: Version) -> Version:
        return parse_version(
            f"{current_version.major}.{current_version.minor +1}.0"
        )

    def next_major_version(self, current_version: Version) -> Version:
        return parse_version(f"{current_version.major + 1}.0.0")

    def next_dev_version(self, current_version: Version) -> Version:
        if current_version.is_devrelease:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }.dev{current_version.dev + 1}"
            )
        elif current_version.is_prerelease:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }{current_version.pre[0]}"
                f"{current_version.pre[1]}+dev1"
            )
        else:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro + 1 }.dev1"
            )

        return release_version

    def next_alpha_version(self, current_version: Version) -> Version:
        if current_version.is_devrelease:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }a1"
            )
        elif current_version.is_prerelease and current_version.pre[0] == "a":
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }a{current_version.pre[1] + 1}"
            )
        else:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro + 1 }a1"
            )

        return release_version

    def next_beta_version(self, current_version: Version) -> Version:
        if current_version.is_devrelease or (
            current_version.is_prerelease and current_version.pre[0] == "a"
        ):
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }b1"
            )
        elif current_version.is_prerelease and current_version.pre[0] == "b":
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }b{current_version.pre[1] + 1}"
            )
        else:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro + 1 }b1"
            )

        return release_version

    def next_release_candidate_version(
        self, current_version: Version
    ) -> Version:
        if current_version.is_devrelease or (
            current_version.is_prerelease and current_version.pre[0] != "rc"
        ):
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }rc1"
            )
        elif current_version.is_prerelease and current_version.pre[0] == "rc":
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro }rc{current_version.pre[1] + 1}"
            )
        else:
            release_version = parse_version(
                f"{current_version.major}."
                f"{current_version.minor}."
                f"{current_version.micro + 1 }rc1"
            )

        return release_version


class VersionCommand(ABC):
    """Generic class usable to implement the
    version commands for several programming languages"""

    project_file_name: str
    version_calculator_class: Type[VersionCalculator] = VersionCalculator

    def __init__(self) -> None:
        self.project_file_path = Path.cwd() / self.project_file_name

    @abstractmethod
    def get_current_version(self) -> Version:
        """Get the current version of this project"""

    @abstractmethod
    def verify_version(
        self, version: Union[Literal["current"], Version]
    ) -> None:
        """Verify the current version of this project"""

    @abstractmethod
    def update_version(
        self, new_version: Version, *, force: bool = False
    ) -> VersionUpdate:
        """Update the current version of this project"""

    def project_found(self) -> bool:
        """
        Returns True if a command has detected a corresponding project
        """
        return self.project_file_path.exists()

    def get_version_calculator(self) -> VersionCalculator:
        return self.version_calculator_class()
